#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef SunOS
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <utime.h>

#define SECN	0
#define SECG	1
#define SECB	2
#define SECQ	3
#define SECZ	4

void nadd(), nfree(), rm(), usage();
char *strrchr(), *strcpy(), *malloc();

int force=0, inq=0, rec=0, oflg=0, sec=SECN, noexec=0, verbose=0;
char *argv0;
char snbuf[BUFSIZ], rnbuf[BUFSIZ];

main(argc, argv)	/* rm: remove files */
char **argv;
{
	extern int optind;
	int c;

	if (argv0=strrchr(argv[0], '/'))
		++argv0;
	else
		argv0=argv[0];

	while ((c=getopt(argc, argv, "finrvosgqz")) != EOF)
		switch (c) {
		case 'f': /* forcably */
			force=1;
			break;
		case 'i': /* inquiry */
			inq=1;
			break;
		case 'n': /* no deleting */
			noexec=1;
			break;
		case 'r': /* recursive */
			rec=1;
			break;
		case 'o': /* overwrite only */
			oflg=1;
			break;
		case 'g': /* Gutmann secure */
			sec=SECG;
			break;
		case 's': /* secure - Bryan's */
			sec=SECB;
			break;
		case 'q': /* secure - quicker */
			sec=SECQ;
			break;
		case 'z': /* zero first */
			sec=SECZ;
			break;
		case 'v': /* verbose */
			verbose++;
			break;
		default: /* others */
			usage();
		}

	if (argc == 1)
		usage();

	if (sec != SECN && sec != SECZ)
		srand(getpid());

	while (optind < argc)
		rm(argv[optind++]);
	exit(0);
}

struct names {
	char *n_nam;
	unsigned n_mod;
	struct names *n_nxt;
} *nroot=NULL, *nlast=NULL;

void
nadd(d, f)
char *d, *f;
{
	struct names *ntmp;

	if ((ntmp=(struct names *)malloc(sizeof(struct names))) == NULL || (ntmp->n_nam=malloc((unsigned)strlen(d)+(unsigned)strlen(f)+2)) == NULL) {
		fprintf(stderr, "%s: Out of memory\n", argv0);
		exit(1);
	}
	sprintf(ntmp->n_nam, "%s/%s", d, f);
	ntmp->n_nxt=NULL;
	if (nlast == NULL)
		nroot=ntmp;
	else
		nlast->n_nxt=ntmp;
	nlast=ntmp;
}

void
nfree()
{
	while (nroot != NULL) {
		nlast=nroot->n_nxt;
		free(nroot->n_nam);
		free((char *)nroot);
		nroot=nlast;
	}
}

char buf[BUFSIZ];
struct stat st;

void
rm(s)
char *s;
{
	int ask;
	int rcnt;
	int err;
	char *p;
	struct names *nrsav, *ntmp;
	int swr;
	char *sn;
	DIR *dp;
#ifdef SunOS
	struct direct *rdp;
#else
	struct dirent *rdp;
#endif

	if (lstat(s, &st) == -1) {
		if (!force)
			fprintf(stderr, "%s: %s not found\n", argv0, s);
		return;
	}

	if ((st.st_mode & S_IFMT) == S_IFDIR) {
		if (verbose > 1)
			fprintf(stderr, "%s: %s is a directory - recursing\n", argv0, s);
		if (rec) {
			if ((dp=opendir(s)) != NULL) {
				while ((rdp=readdir(dp)) != NULL) {
					p=rdp->d_name;
					if (p[0] != '.' || p[1] != '\0' && (p[1] != '.' || p[2] != '\0'))
						nadd(s, p);
				}
				closedir(dp);
				nrsav=nroot;
				nroot=NULL;
				nlast=NULL;
				for (ntmp=nrsav ; ntmp != NULL ; ntmp=ntmp->n_nxt)
					rm(ntmp->n_nam);
				nroot=nrsav;
				nfree();
			}
			if (oflg == 1)
				return;

			ask=inq;
			if (!inq && !force && (st.st_mode & S_IFMT) != S_IFLNK && access(s, 2) == -1) {
				if (noexec)
					ask=1;
				else {
					printf("%s: %s mode %o?", argv0, s, st.st_mode & 0777);
					if (fgets(buf, sizeof(buf), stdin) == NULL)
						exit(2);
					if (*buf != 'y' && *buf != 'Y')
						return;
					chmod(s, st.st_mode | S_IWUSR);
				}
			}

			if (check(s))
				if (noexec)
					printf("%s: !%s %s/\n", argv0, ask?"?":"", s);
				else {
					sn=s;
					if (sec != SECN && sec != SECZ) {
						/*
						 * rename directory
						 */
						for (rcnt=0; rcnt < 16; rcnt++) {
							if (findfile(sn) == -1)
								break;
							if (rename(sn, snbuf) == -1)
								break;
							sync();
							strcpy(rnbuf, snbuf);
							sn=rnbuf;
						}
					}
					if (rmdir(sn) == -1) {
						err=errno;
						if (sn != s)
							rename(sn, s);
						if (!force) {
							errno=err;
							sprintf(buf, "%s: %s/", argv0, s);
							perror(buf);
						}
					} else {
						if (verbose) {
							if (sec != SECN && sec != SECZ) {
								if (sn == s)
									sn="not securely, but still removed";
								else
									sn="securely removed";
							} else if (sec == SECZ)
								sn="zero and removed";
							else
								sn="removed";
							printf("%s: %s %s\n", argv0, s, sn);
						}
						if (sec != SECN && sec != SECZ)
							sync();
					}
				}
		} else if (!force)
			fprintf(stderr, "%s: %s directory\n", argv0, s);
		return;
	}

	ask=inq;
	if (!inq && !force && (st.st_mode & S_IFMT) != S_IFLNK && access(s, 2) == -1) {
		if (noexec)
			ask=1;
		else {
			printf("%s: %s mode %o?", argv0, s, st.st_mode & 0777);
			if (fgets(buf, sizeof(buf), stdin) == NULL)
				exit(2);
			if (*buf != 'y' && *buf != 'Y')
				return;
			chmod(s, st.st_mode | S_IWUSR);
		}
	}

	if (check(s))
		if (noexec)
			printf("%s: !%s %s\n", argv0, ask?"?":"", s);
		else
			erase(s);
}

check(s)
char *s;
{
	if (!noexec && inq) {
		printf("%s: %s?", argv0, s);
		if (fgets(buf, sizeof(buf), stdin) == NULL)
			exit(2);
		if (*buf != 'y' && *buf != 'Y')
			return 0;
	}
	return 1;
}

/*
 * find rename filename
 */
findfile(s)
char *s;
{
	char *ss, *sp;
	int retries=16;

	strcpy(snbuf, s);
	if (ss=strrchr(snbuf, '/'))
		ss++;
	else
		ss=snbuf;

	while (retries--) {
		for (sp=ss; *sp; sp++)
			*sp=randch();
		if (access(snbuf, 0) == -1)
			return 0;
	}
	return -1;
}

randch()
{
	int ch;

	ch=rand()%62;
	if (ch < 10)
		return ch+'0';
	ch -= 10;
	if (ch < 26)
		return ch+'a';
	ch -= 26;
	return ch+'A';
}

/*
 * -1 = 0x92 0x49 0x24
 * -2 = 0x49 0x24 0x92
 * -3 = 0x24 0x92 0x49
 * -4 = 0x6D 0xB6 0xDB
 * -5 = 0xB6 0xDB 0x6D
 * -6 = 0xDB 0x6D 0xB6
 * -51 = random
 * -99 END
 */
int patg[]={	/* official Gutmann 35 pass */
	-51, -51, -51, -51,
	0x55, 0xaa,
	-1, -2, -3,
	0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
	0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,
	-1, -2, -3,
	-4, -5, -6,
	-51, -51, -51, -51,
	-99
};
int patb[]={	/* Bryan Dongrays 30 pass, shorter than Gutman */
	0xff,
	-51,
	0x55, 0xaa,
	-1, -2, -3,
/* FYI: this sequence goes through all 4 bit patterns changing 3 or 4 bits everytime */
	0x77, 0x88, 0x66, 0x99, 0x44, 0xbb, 0x55, 0xaa,
	0xdd, 0x22, 0xcc, 0x33, 0xee, 0x11, 0xff, 0x00,
	-4, -5, -6,
	0xaa, 0x55,
	-51,
	0x00,
	-99
};
int patq[]={	/* quicker 25 pass, only once of each type */
	-51,
	0x55, 0xaa,
	-1, -2, -3,
	-4, -5, -6,
	0x77, 0x88, 0x66, 0x99, 0x44, 0xbb, 0x55, 0xaa,
	0xdd, 0x22, 0xcc, 0x33, 0xee, 0x11, 0xff, 0x00,
	-99
};
int patz[]={	/* Only zeros */
	0x00,
	-99
};

int seqp[]={	/* replacing 0xffff to 0x0000 stepped 0x1111, so more bits change on each step */
	0x7777, 0x8888, 0x6666, 0x9999,
	0x4444, 0xbbbb, 0x5555, 0xaaaa,
	0xdddd, 0x2222, 0xcccc, 0x3333,
	0xeeee, 0x1111, 0xffff, 0x0000,
	-99
};

erase(s)
char *s;
{
	int sok=0;
	int sfd;
	off_t sz;
	int swr;
	int err;
	struct utimbuf ut;
	int rcnt;
	char *sn;
	int *ppat, *pats;

	if (sec != SECN) {
		switch (sec) {
		case SECG:	/* Gutmann */
			pats=patg;
			break;
		case SECB:	/* Bryan's */
			pats=patb;
			break;
		case SECQ:	/* Quicker */
			pats=patq;
			break;
		case SECZ:	/* Only 0s */
			pats=patz;
			break;
		}
		if (verbose > 1)
			fprintf(stderr, "%s: %s%s", argv0, s, sec == SECZ?"":" secure");
		err=0;
		if (st.st_nlink > 1 && (st.st_mode & S_IFMT) == S_IFREG) {
			if (verbose > 1)
				fprintf(stderr, " %d links (no overwrite)", st.st_nlink);
		}
sopen:
		if (st.st_nlink == 1 && (st.st_mode & S_IFMT) == S_IFREG) {
			if ((sfd=open(s, O_WRONLY|O_SYNC)) == -1) {
				if (!force) {
					if (verbose > 1) {
						err=errno;
						fprintf(stderr, "\n");
						errno=err;
					}
					sprintf(buf, "%s: open %s", argv0, s);
					perror(buf);
					return -1;
				}
				if (err == 0) {
					/*
					 * using -f but open failed, since
					 * first time - try setting it writable
					 */
					err=chmod(s, st.st_mode | S_IWUSR);
					if (verbose > 1)
						fprintf(stderr, " %schmod+w", (err == -1)?"!":"");
					if (err == 0) {
						err=1;
						goto sopen;
					}
				}
			} else {
				if (verbose > 1)
					fprintf(stderr, " overwrite:");
				for (ppat=pats; *ppat != -99; ppat++) {
					switch (*ppat) {
					case -51:
						if (verbose > 1)
							fprintf(stderr, " random");
						break;
					case -1:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0x92;
								break;
							case 1:
								buf[swr]=0x49;
								break;
							case 2:
								buf[swr]=0x24;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0x924924", *ppat);
						break;
					case -2:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0x49;
								break;
							case 1:
								buf[swr]=0x24;
								break;
							case 2:
								buf[swr]=0x92;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0x492492", *ppat);
						break;
					case -3:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0x24;
								break;
							case 1:
								buf[swr]=0x92;
								break;
							case 2:
								buf[swr]=0x49;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0x249249", *ppat);
						break;
					case -4:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0x6d;
								break;
							case 1:
								buf[swr]=0xb6;
								break;
							case 2:
								buf[swr]=0xdb;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0x6db6db", *ppat);
						break;
					case -5:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0xb6;
								break;
							case 1:
								buf[swr]=0xdb;
								break;
							case 2:
								buf[swr]=0x6d;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0xb6db6d", *ppat);
						break;
					case -6:
						for (swr=sizeof (buf); swr-- ;)
							switch (swr%3) {
							case 0:
								buf[swr]=0xdb;
								break;
							case 1:
								buf[swr]=0x6d;
								break;
							case 2:
								buf[swr]=0xb6;
								break;
							}
						if (verbose > 1)
							fprintf(stderr, " 0xdb6db6", *ppat);
						break;
					default:
						if (verbose > 1)
							fprintf(stderr, " 0x%02X", *ppat);
						memset(buf, *ppat, sizeof (buf));
						break;
					}
					lseek(sfd, 0L, 0);
					sz=st.st_size;
					while (sz) {
						/*
						 * If random, randomize buffer every time round the loop
						 */
						if (*ppat == -51)
							for (swr=(sz > sizeof (buf))?sizeof (buf):sz ; swr-- ;)
								buf[swr]=rand();
						if ((swr=write(sfd, buf, (sz > sizeof (buf))?sizeof (buf):sz)) == -1) {
							if (!force) {
								if (verbose > 1)
									fprintf(stderr, "\n");
								sprintf(buf, "%s: write %s", argv0, s);
								perror(buf);
								close(sfd);
								return -1;
							}
							close(sfd);
							goto noverwrite;
						}
						sz -= swr;
					}
					sync();
				}
				close(sfd);

				if (oflg == 0 && sec != SECZ) {
					/*
					 * trunctate
					 */
					if (truncate(s, 0) == -1) {
						if (!force) {
							if (verbose > 1)
								fprintf(stderr, "\n");
							sprintf(buf, "%s: truncate %s", argv0, s);
							perror(buf);
							return -1;
						}
						if (verbose > 1)
							fprintf(stderr, " !truncated");
					} else
						if (verbose > 1)
							fprintf(stderr, " truncated");
					sync();
				}

				sok=1;
			}
		} else
			sok=1;

noverwrite:
		sn=s;
		if (sec != SECZ && oflg == 0) {
			/*
			 * rename file
			 */
			if (verbose > 1)
				fprintf(stderr, " rename");
			for (rcnt=0; rcnt < 16; rcnt++) {
				if (findfile(sn) == -1) {
					if (verbose > 1)
						fprintf(stderr, "!");
					break;
				}
				if (rename(sn, snbuf) == -1) {
					if (verbose > 1)
						fprintf(stderr, "!");
					break;
				}
				sync();
				strcpy(rnbuf, snbuf);
				sn=rnbuf;
				if (verbose > 1)
					fprintf(stderr, "+");
			}

			if (st.st_nlink == 1 && (st.st_mode & S_IFMT) != S_IFLNK) {
				/*
				 * splat over file times
				 */
				err=utime(sn, &ut);
				if (verbose > 1)
					fprintf(stderr, " %sutime", (err == -1)?"!":"");
				if (err == 0)
					for (pats=seqp; *pats > 0 ; pats++) {
						sfd = *pats;
						ut.actime=sfd;
						ut.modtime=sfd;
						err=utime(sn, &ut);
						if (verbose > 1)
							fprintf(stderr, "%s", (err == -1)?"!":"+");
						if (err == -1)
							break;
						sync();
					}

				/*
				 * blat out permissions
				 */
				err=chmod(sn, 0);
				if (verbose > 1)
					fprintf(stderr, " %schmod", (err == -1)?"!":"");
				if (err == 0)
					for (pats=seqp; *pats != -99 ; pats++) {
						sfd = *pats;
						err=chmod(sn, sfd);
						if (verbose > 1)
							fprintf(stderr, "%s", (err == -1)?"!":"+");
						if (err == -1)
							break;
						sync();
					}

				/*
				 * alter chown/chgrp
				 */
				err=chown(sn, 0, 0);	/* ignore error */
				if (verbose > 1)
					fprintf(stderr, " %schown", (err == -1)?"!":"");
			}
		}
		if (verbose > 1)
			fprintf(stderr, "\n");
	} else
		sn=s;

	if (oflg == 0) {
		if (unlink(sn) == -1) {
			if (!force) {
				sprintf(buf, "%s: %s", argv0, s);
				perror(buf);
			}
			if (sn != s)
				rename(sn, s);
			return -1;
		}
	}
	if (verbose) {
		if (sec != SECN && sec != SECZ) {
			if (sok)
				sn=(oflg?"securely overwritten":"securely removed");
			else
				sn=(oflg?"not securely overwritten":"not securely, but still removed");
		} else if (sec == SECZ)
			sn=(oflg?"zero'd":"zero and removed");
		else
			sn=(oflg?"overwritten":"removed");
		printf("%s: %s %s\n", argv0, s, sn);
	}

	if (sec != SECN && sec != SECZ)
		sync();
	return 0;
}

#ifdef NEEDRMDIR
rmdir(s)
char *s;
{
	int pid, i, es;
	char *rmd="/bin/rmdir";
	unsigned sleep();

	i=0;
	while ((pid=fork()) == -1)
		if (++i == 10)
			return -1;
		else
			sleep((unsigned)i);
	if (pid == 0) {
		close(0);
		close(1);
		close(2);
		execl(rmd, rmd, s, 0);
		fprintf(stderr, "%s: cannot exec %d\n", argv0, rmd);
		exit(1);
	}
	while ((i=wait(&es)) == -1 && i != pid)
		;
	return es?-1:0;
}
#endif

void
usage()
{
	fprintf(stderr, "Usage: %s [-finrvosgqz] <file/directory>...\n", argv0);
	fprintf(stderr, "\t-f\tforce - report no errors\n");
	fprintf(stderr, "\t-i\tinquire - ask for confirmation\n");
	fprintf(stderr, "\t-n\tno remove actually performed (is verbose)\n");
	fprintf(stderr, "\t-r\trecursively traverse directories\n");
	fprintf(stderr, "\t-v\tverbose. -vv for very verbose\n");
	fprintf(stderr, "\t-o\toverwrite only (don't delete) - use option below\n");
	fprintf(stderr, "\t-z\tzero out file before remove\n");
	fprintf(stderr, "\t-s\tsecurely (can be slow):\n");
	fprintf(stderr, "\t\t\toverwrite multiple patterns (chmod+w if necessary)\n");
	fprintf(stderr, "\t\t\ttruncate rename utime's chmod's chown (if possible)\n");
	fprintf(stderr, "\t\t\tdefault to use 30 patterns\n");
	fprintf(stderr, "\t-g\tsecurely with Gutman 35 pass\n");
	fprintf(stderr, "\t-q\tsecurely with 25 passes\n");
	exit(1);
}
